#include "xbasic_types.h"
#include "xio.h"
#include "xintc_l.h"
#include "xparameters.h"
#include "xuartlite_l.h"

#define R   0xE0 //red
#define G   0x1C //green
#define B   0x02 //blue
#define Y   0xFC //yellow
#define V   0xE2 //violet
#define W   0xFE //white
#define FLOOR   0xAD
#define CEILING 0x1F

//sine/cosines scaled to 11 bits
#define ISINVALUE   0x00000132
#define ICOSVALUE   0x000007E9
#define ISINNVALUE  0xFFFFFECE
#define ICOSNVALUE  0x000007E9

#define SCREENWIDTH 640
#define VGA_START 0x00800000
#define HEIGHT 480
#define HEIGHTMINUS1 479
#define SCALEDHEIGHT 15360
#define HALFHEIGHT 240

#define STARTADDR XPAR_VGA_BASEADDR
//#define STOPADDR  STARTADDR + 1024
//#define COLORADDR COLORADDR + 1024



extern void XIntc_DefaultHandler (void *);
extern void Xhw_InterruptAHandler (void *);


extern int uart_interrupt_count;
extern void microblaze_enable_icache();
extern void microblaze_enable_interrupts();
extern void microblaze_disable_interrupts();
//extern void print(char *);
inline extern int divsi3_proc(int, int);


int uart_interrupt_count = 0;
char uart_character;

//a circular buffer is sized to hold input data
#define BUFSIZE 2 //input character buffer size
char buffer[BUFSIZE+1];
char *lastin  = buffer;
char *lastout = buffer;

/*
 * Interrupt service routine for the UART
 */
void uart_handler(void *callback)
{
  Xuint32 IsrStatus;
  Xuint8 incoming_character;

  /* Check the ISR status register so we can identify the interrupt source */
  IsrStatus = XIo_In32(XPAR_MYUART_BASEADDR + XUL_STATUS_REG_OFFSET);
  
  if ((IsrStatus & (XUL_SR_RX_FIFO_FULL | XUL_SR_RX_FIFO_VALID_DATA)) != 0) {
    /* The input FIFO contains data: read it */
    incoming_character =
      (Xuint8) XIo_In32( XPAR_MYUART_BASEADDR + XUL_RX_FIFO_OFFSET );

    //add incoming character to buffer
    if (lastin == buffer + BUFSIZE)
      //rollover to beginning of buffer if necessary
      lastin=buffer;
    *lastin = incoming_character;
    lastin++;

    //uart_character = incoming_character;
    //++uart_interrupt_count;
  }
  if ((IsrStatus & XUL_SR_TX_FIFO_EMPTY) != 0) {
    /* The output FIFO is empty: we can send another character */
  }   
}







int main()
{
    Xuint8 worldMap[16][16]= {{W,W,W,W,W,W,W,W,W,W,W,Y,Y,Y,Y,Y},
			      {W,0,0,B,B,0,0,0,0,0,0,0,0,0,0,Y},
			      {W,0,0,0,G,G,G,G,G,0,0,0,0,0,0,Y},
			      {W,0,0,0,0,0,0,0,0,0,0,0,0,0,0,Y},
			      {W,R,R,R,R,R,R,R,R,R,0,0,0,0,0,Y},
			      {W,0,0,0,0,0,G,0,0,R,0,0,0,0,0,Y},
			      {W,0,G,G,G,G,G,0,0,R,0,0,0,0,0,Y},
			      {W,0,B,0,0,Y,0,0,V,R,G,G,G,0,0,Y},
			      {W,0,B,0,0,Y,0,0,V,1,1,0,B,0,0,W},
			      {W,0,B,0,0,Y,0,0,V,0,0,0,B,0,0,W},
			      {W,0,B,0,0,Y,0,0,V,0,0,0,B,0,0,W},
			      {W,0,B,0,0,Y,0,0,V,0,0,0,B,0,0,W},
			      {W,0,B,0,0,Y,0,0,V,0,0,0,B,0,0,W},
			      {W,0,0,0,0,0,0,0,V,0,0,0,0,0,0,W},
			      {W,0,0,0,0,0,0,0,V,0,0,0,0,0,0,W},
			      {W,W,W,W,W,W,W,W,W,W,W,W,W,W,W,W}};

    char *end_of_buffer = buffer+2;
    Xuint8 draw=1;
    char c;
    char process;
    
    int posX   =  2<<16, posY   = 2<<16;
    int dirX   = -1<<10, dirY   = 0;
    int planeX = 0,      planeY = (43254)>>6;
    int factor;
    int oldDirX, oldPlaneX;
    int lineHeight, drawStart, drawEnd;
    Xuint8 myColor;
    Xuint32 addr = XPAR_VGA_BASEADDR;
    register int count, oldMapX=0, oldMapY, mapX, mapY, rayDirX, rayDirY;
    register int x,y, temp;

    buffer[2]='\0';
    XIntc_RegisterHandler( XPAR_INTC_BASEADDR, XPAR_MYUART_DEVICE_ID,
                           (XInterruptHandler)uart_handler, (void *)0);
    XIntc_mEnableIntr( XPAR_INTC_BASEADDR, XPAR_MYUART_INTERRUPT_MASK);
    XIntc_mMasterEnable( XPAR_INTC_BASEADDR );
    XIntc_Out32(XPAR_INTC_BASEADDR + XIN_MER_OFFSET, XIN_INT_MASTER_ENABLE_MASK);
    XUartLite_mEnableIntr(XPAR_MYUART_BASEADDR);
    print("Welcome to MAYD\r\n\r\n");
    microblaze_enable_icache();
    microblaze_enable_interrupts();

    
    addr = XPAR_VGA_BASEADDR +800-640 -16;

    for (;;)
    {
	//get next character
	microblaze_disable_interrupts();
	if (lastin != lastout)
	{
	    c=*lastout;      
	    microblaze_enable_interrupts();
	    if (lastout == end_of_buffer)
		lastout = buffer;
	    else
		lastout++;


	    if ( (c == 's') || (c == 'w') || (c == 'a') || (c == 'd'))
	      process=c;
	    else if (c == 32) //space
	      process = 0;
	      

	}
	else
	    microblaze_enable_interrupts();



	    //print("got character: ");
	    //putnum(c);
	    //print("\r\n");


	    if (process == 'w') //up
	    {
		temp = dirX <<2;
		if(worldMap[(posX + temp)>>16][posY>>16] == 0)
		    posX += temp;
		temp = dirY <<2;
		if(worldMap[posX>>16][(posY + temp)>>16] == 0)
		    posY += temp;
		draw=1;
	    }
	    else if (process == 's') //down
	    {
		temp = dirX <<2;
		if(worldMap[(posX - temp)>>16][posY>>16] == 0)
		    posX -= temp;
		temp = dirY <<2;
		if(worldMap[posX>>16][(posY - temp)>>16] == 0)
		   posY -= temp;
		draw=1;
	    }
	    else if (process == 'a') //left
	    {
		oldDirX   = dirX;
		oldPlaneX = planeX;
		dirX   = (int)(dirX      * ICOSVALUE - dirY   * ISINVALUE)>>11;
		dirY   = (int)(oldDirX   * ISINVALUE + dirY   * ICOSVALUE)>>11;
		planeX = (int)(planeX    * ICOSVALUE - planeY * ISINVALUE)>>11;
		planeY = (int)(oldPlaneX * ISINVALUE + planeY * ICOSVALUE)>>11;
		draw=1;
	    }
	    else if (process == 'd') //right
	    {
		oldDirX   = dirX;
		oldPlaneX = planeX;
		dirX   = (int)(dirX      * ICOSNVALUE - dirY   * ISINNVALUE)>>11;
		dirY   = (int)(oldDirX   * ISINNVALUE + dirY   * ICOSNVALUE)>>11;
		planeX = (int)(planeX    * ICOSNVALUE - planeY * ISINNVALUE)>>11;
		planeY = (int)(oldPlaneX * ISINNVALUE + planeY * ICOSNVALUE)>>11;
		draw=1;
	    }
    

	
	if (draw!=1)
	    continue;
	
	//MAYD - MAYD a yellow day
        factor=0xFFFEFFCD;


	addr = XPAR_VGA_BASEADDR +800-640 -16;
	for(x = 0; x < SCREENWIDTH; x++)
	{
	    factor +=0xCD;
	    //initalize
	    count = 0;
	    rayDirX = dirX + ((planeX * factor) >>16);
	    rayDirY = dirY + ((planeY * factor) >>16);
	    mapX = posX;
	    mapY = posY;

	    //raycasting is loads of fun!
	    while ((temp = worldMap[mapX>>16][mapY>>16]) == 0)
	    {
		count++;
		oldMapY = mapY;
		oldMapX = mapX;
		mapX += rayDirX;
		mapY += rayDirY;
	    }

	    //set color
	    myColor = temp;
	    if(worldMap[oldMapX>>16][mapY>>16]) myColor+=1;
	    if(myColor==1) myColor = 0xFF & rayDirX & rayDirY;

	    //cant get this to inline
	    lineHeight = SCALEDHEIGHT / count;

	    temp = lineHeight >>1;
	    drawStart = HALFHEIGHT - temp;
	    drawEnd   = HALFHEIGHT + temp;
	    if(drawStart < 0)		drawStart = 0;
	    if(drawEnd   >= HEIGHT)	drawEnd   = HEIGHTMINUS1;	

	    //do actual drawing
	    XIo_Out8(addr+x,drawStart);
	    XIo_Out8(addr+x+1024,drawEnd-240);
	    XIo_Out8(addr+x+2048,myColor);
	    

	    /*
	    XIo_Out8(addr+x,0x50);
	    XIo_Out8(addr+x+1024,0x80);
	    XIo_Out8(addr+x+2048,myColor);
	    */

	    //	    addr+1;
	}
	draw=0;
    }
    return 0;
}
